@@ -5,26 +5,28 @@
 
 # pyre-unsafe
 
+import glob
 import hashlib
 import json
 import logging
+import multiprocessing
 import os
 import shutil
 import site
+import subprocess
 import sys
 from logging import Logger
+from pathlib import Path
 from typing import Dict, List, Optional, Union
 
-from . import (
+from .exceptions import EnvironmentException
+from .filesystem import assert_readable_directory, expand_relative_path, find_root
+from .find_directories import (
     BINARY_NAME,
     CONFIGURATION_FILE,
     LOCAL_CONFIGURATION_FILE,
-    LOG_DIRECTORY,
-    assert_readable_directory,
     find_typeshed,
-    number_of_workers,
 )
-from .exceptions import EnvironmentException
 
 
 LOG: Logger = logging.getLogger(__name__)
@@ -39,6 +41,35 @@ class SearchPathElement:
         self.root = os.path.expanduser(root)
         self.subdirectory = subdirectory
 
+    @staticmethod
+    def expand(path: Union[Dict[str, str], str]) -> "SearchPathElement":
+        if isinstance(path, str):
+            return SearchPathElement(path)
+        else:
+            if "root" in path and "subdirectory" in path:
+                root = path["root"]
+                subdirectory = path["subdirectory"]
+                return SearchPathElement(root, subdirectory)
+            elif "site-package" in path:
+                site_root = site.getsitepackages()
+                subdirectory = path["site-package"]
+
+                found_element = None
+                for root in site_root:
+                    site_package_element = SearchPathElement(root, subdirectory)
+                    if os.path.isdir(site_package_element.path()):
+                        found_element = site_package_element
+                if found_element is None:
+                    raise InvalidConfiguration(
+                        "Cannot find site package '{}'".format(subdirectory)
+                    )
+                return found_element
+            else:
+                raise InvalidConfiguration(
+                    "Search path elements must have `root` and `subdirectory` "
+                    "specified."
+                )
+
     def path(self) -> str:
         subdirectory = self.subdirectory
         if subdirectory is not None:
@@ -61,34 +92,6 @@ class SearchPathElement:
             return self.root == other.root and self.subdirectory == other.subdirectory
 
 
-def expand_search_path(path: Union[Dict[str, str], str]) -> SearchPathElement:
-    if isinstance(path, str):
-        return SearchPathElement(path)
-    else:
-        if "root" in path and "subdirectory" in path:
-            root = path["root"]
-            subdirectory = path["subdirectory"]
-            return SearchPathElement(root, subdirectory)
-        elif "site-package" in path:
-            site_root = site.getsitepackages()
-            subdirectory = path["site-package"]
-
-            found_element = None
-            for root in site_root:
-                site_package_element = SearchPathElement(root, subdirectory)
-                if os.path.isdir(site_package_element.path()):
-                    found_element = site_package_element
-            if found_element is None:
-                raise InvalidConfiguration(
-                    "Cannot find site package '{}'".format(subdirectory)
-                )
-            return found_element
-        else:
-            raise InvalidConfiguration(
-                "Search path elements must have `root` and `subdirectory` specified."
-            )
-
-
 class _ConfigurationFile:
     def __init__(self, file) -> None:
         self._deprecated = {"do_not_check": "ignore_all_errors"}
@@ -136,36 +139,26 @@ class _ConfigurationFile:
         Return all keys not consumed yet. Some keys are explicitly whitelisted.
         """
         return self._configuration.keys() - {
-            "buck_builder_binary",
             "buck_mode",
-            "continuous",
             "differential",
-            "push_blocking",
-            "pyre_client",
+            "stable_client",
+            "unstable_client",
             "saved_state",
             "taint_models_path",
-            "oncall",
         }
 
 
-def expand_relative_path(root: str, path: str) -> str:
-    path = os.path.expanduser(path)
-    if os.path.isabs(path):
-        return path
-    else:
-        return os.path.join(root, path)
-
-
 class Configuration:
     disabled: bool = False
 
     def __init__(
         self,
-        local_configuration: Optional[str] = None,
+        project_root: str,
+        local_root: Optional[str] = None,
         search_path: Optional[List[str]] = None,
         binary: Optional[str] = None,
         typeshed: Optional[str] = None,
-        preserve_pythonpath: bool = False,
+        buck_builder_binary: Optional[str] = None,
         excludes: Optional[List[str]] = None,
         formatter: Optional[str] = None,
         logger: Optional[str] = None,
@@ -177,7 +170,8 @@ class Configuration:
         self.formatter = formatter
         self.ignore_all_errors = []
         self.number_of_workers: int = 0
-        self.local_configuration: Optional[str] = None
+        self.project_root: str = project_root
+        self._local_root: Optional[str] = local_root
         self.taint_models_path: List[str] = []
         self.file_hash: Optional[str] = None
         self.extensions: List[str] = []
@@ -186,30 +180,16 @@ class Configuration:
         self._version_hash: Optional[str] = None
         self._binary: Optional[str] = None
         self._typeshed: Optional[str] = None
+        self._buck_builder_binary: Optional[str] = None
         self.strict: bool = False
         self._use_buck_builder: Optional[bool] = None
         self.ignore_infer: List[str] = []
 
-        # Handle search path from multiple sources
-        self._search_path = []
-        if preserve_pythonpath:
-            for path in os.getenv("PYTHONPATH", default="").split(":"):
-                if path != "":
-                    if os.path.isdir(path):
-                        self._search_path.append(SearchPathElement(path))
-                    else:
-                        LOG.warning(
-                            "`{}` is not a valid directory, dropping it "
-                            "from PYTHONPATH".format(path)
-                        )
-            # sys.path often includes '' and a zipped python version, so
-            # we don't log warnings for non-dir entries
-            sys_path = [
-                SearchPathElement(path) for path in sys.path if os.path.isdir(path)
-            ]
-            self._search_path.extend(sys_path)
+        self._search_path: List[SearchPathElement] = []
         if search_path:
-            search_path_elements = [expand_search_path(path) for path in search_path]
+            search_path_elements = [
+                SearchPathElement.expand(path) for path in search_path
+            ]
             self._search_path.extend(search_path_elements)
         # We will extend the search path further, with the config file
         # items, inside _read().
@@ -220,23 +200,26 @@ class Configuration:
         if typeshed:
             self._typeshed = typeshed
 
+        if buck_builder_binary:
+            self._buck_builder_binary = buck_builder_binary
+
         self.excludes: List[str] = []
         if excludes:
             self.excludes.extend(excludes)
 
-        if local_configuration:
-            # Handle local configuration explicitly configured on the
-            # commandline.
-            self._check_read_local_configuration(
-                local_configuration, fail_on_error=True
-            )
+        if local_root:
+            self._check_read_local_configuration(local_root)
         if log_directory:
             self.ignore_all_errors.append(log_directory)
 
         self.autocomplete = False
 
+        self.other_critical_files: List[str] = []
+
+        project_configuration = os.path.join(project_root, CONFIGURATION_FILE)
+
         # Order matters. The values will only be updated if a field is None.
-        self._read(CONFIGURATION_FILE)
+        self._read(project_configuration)
         self._override_version_hash()
         self._resolve_versioned_paths()
         self._apply_defaults()
@@ -301,6 +284,15 @@ class Configuration:
                     "the root of the `typeshed` directory.".format(self.typeshed)
                 )
 
+            expanded_ignore_paths = []
+            for path in self.ignore_all_errors:
+                expanded = glob.glob(path)
+                if not expanded:
+                    expanded_ignore_paths.append(path)
+                else:
+                    expanded_ignore_paths += expanded
+            self.ignore_all_errors = expanded_ignore_paths
+
             non_existent_ignore_paths = [
                 path for path in self.ignore_all_errors if not os.path.exists(path)
             ]
@@ -331,30 +323,10 @@ class Configuration:
                     if path not in non_existent_infer_paths
                 ]
 
-            typeshed_subdirectories = os.listdir(self.typeshed)
-            if "stdlib" not in typeshed_subdirectories:
-                LOG.warning(
-                    "Typeshed at `%s` contains subdirectories:\n%s",
-                    self.typeshed,
-                    typeshed_subdirectories,
-                )
-                raise InvalidConfiguration(
-                    "`typeshed` location must contain a `stdlib` directory."
-                )
-
-            for typeshed_subdirectory_name in typeshed_subdirectories:
+            for typeshed_subdirectory_name in ["stdlib", "third_party"]:
                 typeshed_subdirectory = os.path.join(
                     self.typeshed, typeshed_subdirectory_name
                 )
-                if (
-                    not os.path.isdir(typeshed_subdirectory)
-                    or typeshed_subdirectory_name == "tests"
-                    or typeshed_subdirectory_name[0] == "."
-                    or typeshed_subdirectory_name == "__pycache__"
-                ):
-                    # Ignore some well-known directories we do not care about.
-                    continue
-
                 assert_readable_directory(typeshed_subdirectory)
                 for typeshed_version_directory_name in os.listdir(
                     typeshed_subdirectory
@@ -372,7 +344,15 @@ class Configuration:
 
             # Validate elements of the search path.
             for element in self._search_path:
-                assert_readable_directory(element.path())
+                assert_readable_directory(
+                    element.path(),
+                    error_message_prefix="Invalid search_path in configuration: ",
+                )
+
+            if not is_list_of_strings(self.other_critical_files):
+                raise InvalidConfiguration(
+                    "`critical_files` field must be a list of strings."
+                )
         except InvalidConfiguration as error:
             raise EnvironmentException("Invalid configuration: {}".format(str(error)))
 
@@ -394,6 +374,10 @@ class Configuration:
             raise InvalidConfiguration("Configuration invalid: no typeshed specified")
         return typeshed
 
+    @property
+    def buck_builder_binary(self) -> Optional[str]:
+        return self._buck_builder_binary
+
     @property
     def use_buck_builder(self) -> bool:
         return self._use_buck_builder or False
@@ -405,13 +389,8 @@ class Configuration:
         return [element.command_line_argument() for element in self._search_path]
 
     @property
-    def local_configuration_root(self) -> Optional[str]:
-        local_configuration = self.local_configuration
-        if local_configuration:
-            if os.path.isdir(local_configuration):
-                return local_configuration
-            else:
-                return os.path.dirname(local_configuration)
+    def local_root(self) -> Optional[str]:
+        return self._local_root
 
     @property
     def log_directory(self) -> str:
@@ -420,32 +399,80 @@ class Configuration:
             raise InvalidConfiguration("Configuration was not validated")
         return log_directory
 
-    def _check_read_local_configuration(self, path: str, fail_on_error: bool) -> None:
-        if fail_on_error and not os.path.exists(path):
+    def get_binary_version(self) -> Optional[str]:
+        status = subprocess.run(
+            [self.binary, "-version"], stdout=subprocess.PIPE, universal_newlines=True
+        )
+        if status.returncode == 0:
+            return status.stdout.strip()
+        else:
+            return None
+
+    def _check_nested_configurations(self, local_root: str) -> None:
+        # TODO(T67874463): Handle sanity checks against project configurations
+        parent_local_root = find_root(
+            os.path.dirname(local_root.rstrip("/")), LOCAL_CONFIGURATION_FILE
+        )
+        if not parent_local_root:
+            return
+
+        parent_error = None
+        excluded_from_parent = False
+        local_root: Path = Path(local_root).resolve()
+        try:
+            parent_local_configuration = Configuration(
+                project_root=self.project_root,
+                local_root=parent_local_root,
+                search_path=[search_path.root for search_path in self._search_path],
+                binary=self._binary,
+                typeshed=self._typeshed,
+                buck_builder_binary=self._buck_builder_binary,
+                excludes=self.excludes,
+                logger=self.logger,
+                formatter=self.formatter,
+                log_directory=self._log_directory,
+            )
+            paths_to_ignore = list(local_root.parents) + [local_root]
+            for ignore_element in parent_local_configuration.ignore_all_errors:
+                if Path(ignore_element).resolve() in paths_to_ignore:
+                    excluded_from_parent = True
+        except EnvironmentException as error:
+            parent_error = error
+
+        if not excluded_from_parent:
+            relative_path = str(local_root.relative_to(parent_local_root))
+            error_message = (
+                "Local configuration is nested under another local configuration at "
+                f"`{parent_local_root}`.\nPlease add `{relative_path}` to the "
+                "`ignore_all_errors` field of the parent, or combine the sources "
+                "into a single configuration, or split the parent configuration to "
+                "avoid inconsistent errors."
+            )
+            raise EnvironmentException(error_message)
+        elif parent_error:
+            raise EnvironmentException(parent_error)
+
+    def _check_read_local_configuration(self, path: str) -> None:
+        if not os.path.exists(path):
             raise EnvironmentException(
                 "Local configuration path `{}` does not exist.".format(path)
             )
 
-        if os.path.isdir(path):
-            local_configuration = os.path.join(path, CONFIGURATION_FILE + ".local")
+        local_configuration = os.path.join(path, LOCAL_CONFIGURATION_FILE)
+        if not os.path.exists(local_configuration):
+            raise EnvironmentException(
+                "Local configuration directory `{}` does not contain "
+                "a `{}` file.".format(path, LOCAL_CONFIGURATION_FILE)
+            )
 
-            if not os.path.exists(local_configuration):
-                if fail_on_error:
-                    raise EnvironmentException(
-                        "Local configuration directory `{}` does not contain "
-                        "a `{}` file.".format(path, CONFIGURATION_FILE + ".local")
-                    )
-                else:
-                    LOG.debug(
-                        "Configuration will be read from the project root: "
-                        "`{}`".format(os.getcwd())
-                    )
-            else:
-                self.local_configuration = local_configuration
-        else:
-            local_configuration = path
-            self.local_configuration = local_configuration
+        self._check_nested_configurations(path)
         self._read(local_configuration)
+        if not self.source_directories and not self.targets:
+            raise EnvironmentException(
+                "Local configuration `{}` does not specify any sources to type check.".format(
+                    local_configuration
+                )
+            )
 
     def _read(self, path: str) -> None:
         try:
@@ -493,8 +520,7 @@ class Configuration:
                 ignore_all_errors = configuration.consume(
                     "ignore_all_errors", default=[]
                 )
-                # Deprecated.
-                ignore_all_errors += configuration.consume("do_not_check", default=[])
+
                 configuration_path = os.path.dirname(os.path.realpath(path))
                 self.ignore_all_errors.extend(
                     [
@@ -523,13 +549,24 @@ class Configuration:
                     binary = expand_relative_path(configuration_path, binary)
                 self._binary = binary
 
+                buck_builder_binary = configuration.consume(
+                    "buck_builder_binary", current=self._buck_builder_binary
+                )
+                if buck_builder_binary is not None:
+                    self._buck_builder_binary = expand_relative_path(
+                        root=configuration_path, path=buck_builder_binary
+                    )
+
                 additional_search_path = configuration.consume(
                     "search_path", default=[]
                 )
 
                 if isinstance(additional_search_path, list):
                     self._search_path.extend(
-                        [expand_search_path(path) for path in additional_search_path]
+                        [
+                            SearchPathElement.expand(path)
+                            for path in additional_search_path
+                        ]
                     )
                 else:
                     self._search_path.append(SearchPathElement(additional_search_path))
@@ -584,6 +621,14 @@ class Configuration:
 
                 self.autocomplete = configuration.consume("autocomplete", default=False)
 
+                self.other_critical_files = configuration.consume(
+                    "critical_files", default=[]
+                )
+
+                # Warn on deprecated fields.
+                for deprecated_field in configuration._deprecated.keys():
+                    configuration.consume(deprecated_field)
+
                 # This block should be at the bottom to be effective.
                 unused_keys = configuration.unused_keys()
                 if unused_keys:
@@ -639,7 +684,10 @@ class Configuration:
                 LOG.info("Found: `%s`", self._binary)
 
         if self.number_of_workers == 0:
-            self.number_of_workers = number_of_workers()
+            try:
+                self.number_of_workers = max(multiprocessing.cpu_count() - 4, 1)
+            except NotImplementedError:
+                self.number_of_workers = 4
 
         if not self._typeshed:
             LOG.info("No typeshed specified, looking for it")
