@@ -6,9 +6,11 @@
 import enum
 import logging
 from collections import namedtuple
-from itertools import islice, tee
+from itertools import islice
 from typing import Any, Dict, List, Optional, Set, Tuple, Type
 
+from graphene import relay
+from graphene_sqlalchemy import SQLAlchemyObjectType
 from munch import Munch
 from sqlalchemy import (
     Boolean,
@@ -30,10 +32,23 @@ from sqlalchemy import (
 )
 from sqlalchemy.dialects import mysql, sqlite
 from sqlalchemy.dialects.mysql import BIGINT, INTEGER
+from sqlalchemy.exc import NoSuchTableError
 from sqlalchemy.ext.associationproxy import association_proxy
 from sqlalchemy.ext.declarative import declarative_base
 from sqlalchemy.orm import Session, relationship
 
+from .db import DB
+from .db_support import (
+    DBID,
+    BIGDBIDType,
+    DBIDType,
+    MutableRecordMixin,
+    PrepareMixin,
+    PrimaryKeyBase,
+    PrimaryKeyGeneratorBase,
+    RecordMixin,
+)
+from .decorators import classproperty
 from .errors import AIException
 from .iterutil import split_every
 
@@ -41,24 +56,12 @@ from .iterutil import split_every
 log = logging.getLogger("sapp")
 
 
-# For use on enums to alias upper case value.
-#
-# FLAKE8 does not understand that this is a static property
-# flake8: noqa B902
-class classproperty(property):
-    def __get__(self, cls, owner):
-        return classmethod(self.fget).__get__(None, owner)()
-
-
 Base = declarative_base()
 INNODB_MAX_INDEX_LENGTH = 767
 HANDLE_LENGTH = 255
 MESSAGE_LENGTH = 4096
 SHARED_TEXT_LENGTH = 4096
 
-"""Number of variables that can safely be set on a single DB call"""
-BATCH_SIZE = 450
-
 """Models used to represent DB entries
 
 An Issue is a particular problem found. It can exist across multiple commits.  A
@@ -68,157 +71,6 @@ IssueInstances have per run information, like source location, while Issues have
 attributes like the status of an issue.
 """
 
-"""Tables that should be removed from existing databases.
-
-These are tables that are NO LONGER in use, and will be removed. Don't add
-tables here until all code that was using them has been updated.
-"""
-PurgeMetadata = MetaData()
-for t in [
-    "active_jobs",
-    "deletion_history",
-    "herald_task_history",
-    "precondition_incremental_lookup",
-    "run_precondition_assoc",
-    "run_postcondition_assoc",
-]:
-    Table(t, PurgeMetadata)
-
-
-class PrepareMixin(object):
-    @classmethod
-    def prepare(cls, session, pkgen, items):
-        """This is called immediately before the items are written to the
-        database. pkgen is passed in to allow last-minute resolving of ids.
-        """
-        for item in cls.merge(session, items):
-            if hasattr(item, "id"):
-                item.id.resolve(id=pkgen.get(cls), is_new=True)
-            yield cls.to_dict(item)
-
-    @classmethod
-    def merge(cls, session, items):
-        """Models should override this to perform a merge"""
-        return items
-
-    @classmethod
-    def _merge_by_key(cls, session, items, attr):
-        return cls._merge_by_keys(
-            session, items, lambda item: getattr(item, attr.key), attr
-        )
-
-    @classmethod
-    def _merge_by_keys(cls, session, items, hash_item, *attrs):
-        """An object can have multiple attributes as its key. This merges the
-        items to be added with existing items in the database based on their
-        key(s).
-
-        session: Session object for querying the DB.
-        items: Iterator of items to be added to the DB.
-        hash_item: Function that takes as in put the item to be added and
-                   returns a hash of it.
-        attrs: List of attributes of the object/class that represent the
-               object's key.
-
-        Returns the next item (in items) that is not already in the DB.
-        """
-        # Note: items is an iterator, not an iterable, 'tee' is a must.
-        items_iter1, items_iter2 = tee(items)
-
-        keys = {}  # map of hash -> keys of the item
-        for i in items_iter1:
-            # An item's key is a map of 'attr -> item[attr]' where attr is
-            # usually a column name.
-            # For 'SharedText', its key would look like: {
-            #   "kind": "feature",
-            #   "contents": "via tito",
-            # }
-            item_hash = hash_item(i)
-            keys[item_hash] = {attr.key: getattr(i, attr.key) for attr in attrs}
-
-        # Find existing items.
-        existing_ids = {}  # map of item_hash -> existing ID
-        cls_attrs = [getattr(cls, attr.key) for attr in attrs]
-        for fetch_keys in split_every(BATCH_SIZE, keys.values()):
-            filters = []
-            for fetch_key in fetch_keys:
-                # Sub-filters for checking if item with fetch_key is in the DB
-                # Example: [
-                #   SharedText.kind.__eq__("feature"),
-                #   SharedText.contents.__eq__("via tito"),
-                # ]
-                subfilter = [
-                    getattr(cls, attr).__eq__(val) for attr, val in fetch_key.items()
-                ]
-                filters.append(and_(*subfilter))
-            existing_items = (
-                session.query(cls.id, *cls_attrs).filter(or_(*(filters))).all()
-            )
-            for existing_item in existing_items:
-                item_hash = hash_item(existing_item)
-                existing_ids[item_hash] = existing_item.id
-
-        # Now see if we can merge
-        new_items = {}
-        for i in items_iter2:
-            item_hash = hash_item(i)
-            if item_hash in existing_ids:
-                # The key is already in the DB
-                i.id.resolve(existing_ids[item_hash], is_new=False)
-            elif item_hash in new_items:
-                # The key is already in the list of new items
-                i.id.resolve(new_items[item_hash].id, is_new=False)
-            else:
-                # The key is new
-                new_items[item_hash] = i
-                yield i
-
-    @classmethod
-    def _merge_assocs(cls, session, items, id1, id2):
-        new_items = {}
-        for i in items:
-            r1 = getattr(i, id1.key)
-            r2 = getattr(i, id2.key)
-            key = (r1.resolved(), r2.resolved())
-            if key not in new_items:
-                new_items[key] = i
-                yield i
-
-
-# The record mixin class is more efficient than the MutableRecordMixin, so it
-# should be preferred. But the performance isn't from the mutability, it's
-# because we use namedtuples, which creates a new class on demand, which uses
-# __slots__, which is more efficient. Both of these mixins can be replaced when
-# we have dynamically created classes with the slots set. But until then,
-# prefer RecordMixin unless you need to change fields after creation.
-class RecordMixin(object):
-    _record = None
-
-    @classmethod
-    def Record(cls, extra_fields=None, **kwargs):
-        if not cls._record:
-            if not extra_fields:
-                extra_fields = []
-            mapper = inspect(cls)
-            keys = [c.key for c in mapper.column_attrs] + ["model"] + extra_fields
-            cls._record = namedtuple(cls.__name__ + "Record", keys)
-
-        return cls._record(model=cls, **kwargs)
-
-    @classmethod
-    def to_dict(cls, obj):
-        return obj._asdict()
-
-
-class MutableRecordMixin(object):
-    @classmethod
-    def Record(cls, **kwargs):
-        return Munch(model=cls, **kwargs)
-
-    @classmethod
-    def to_dict(cls, obj):
-        return obj.toDict()
-
 
 class SourceLocation(object):
     """The location in a source file that an error occurred in
@@ -327,107 +179,6 @@ class SourceLocationsType(types.TypeDecorator):
         return [SourceLocation.from_string(location) for location in locations]
 
 
-# The following three DBID classes require some explanation. Normally models
-# will reference each other by their id. But we do bulk insertion at the end
-# of our processing, which means the id isn't set until later. Having a DBID
-# object allows these models to reference each other before that point. When
-# we are ready to insert into the database, PrimaryKeyGenerator will give it
-# an ID. Any other models referencing that DBID object will now be able to use
-# the real id.
-
-
-class DBID(object):
-    __slots__ = ["_id", "is_new", "local_id"]
-
-    # Temporary IDs that are local per run (local_id) are assigned for each
-    # DBID object on creation. This acts as a key for the object in map-like
-    # structures of DB objects without having to define a hashing function for
-    # each of them. next_id tracks the next available int to act as an id.
-    next_id: int = 0
-
-    def __init__(self, id=None):
-        self.resolve(id)
-        self.local_id: int = DBID.next_id
-        DBID.next_id += 1
-
-    def resolve(self, id, is_new=True):
-        self._check_type(id)
-        self._id = id
-        self.is_new = is_new
-        return self
-
-    def resolved(self):
-        id = self._id
-
-        # We allow one level of a DBID pointing to another DBID
-        if isinstance(id, DBID):
-            id = id.resolved()
-
-        return id
-
-    def _check_type(self, id):
-        if not isinstance(id, (int, type(None), DBID)):
-            raise TypeError(
-                "id expected to be type '{}' but was type '{}'".format(int, type(id))
-            )
-
-    # Allow DBIDs to be added and compared as ints
-    def __int__(self):
-        return self.resolved()
-
-    def __str__(self):
-        return str(self.resolved())
-
-    def __add__(self, other):
-        return int(self) + int(other)
-
-    def __lt__(self, other):
-        return int(self) < int(other)
-
-    def __gt__(self, other):
-        return int(self) > int(other)
-
-    def __ge__(self, other):
-        return int(self) >= int(other)
-
-    def __le__(self, other):
-        return int(self) <= int(other)
-
-    def __repr__(self):
-        return "<{}(id={}) object at 0x{:x}>".format(
-            self.__class__.__name__, self._id, id(self)
-        )
-
-
-class DBIDType(types.TypeDecorator):
-    impl = types.Integer
-
-    def process_bind_param(self, value, dialect):
-        # If it is a DBID wrapper, then write the contained value. Otherwise it
-        # may be resolved already, or None.
-        if isinstance(value, DBID):
-            return value.resolved()
-        else:
-            return value
-
-    def process_result_value(self, value, dialect):
-        return DBID(value)
-
-    def load_dialect_impl(self, dialect):
-        if dialect.name == "mysql":
-            return dialect.type_descriptor(mysql.INTEGER(unsigned=True))
-        return self.impl
-
-
-class BIGDBIDType(DBIDType):
-    impl = types.BigInteger
-
-    def load_dialect_impl(self, dialect):
-        if dialect.name == "mysql":
-            return dialect.type_descriptor(mysql.BIGINT(unsigned=True))
-        return self.impl
-
-
 # See Issue.merge for information about replace_assocs
 
 
@@ -449,7 +200,6 @@ class IssueBIGDBIDType(BIGDBIDType):
         return IssueDBID(value)
 
 
-# pyre-fixme[11]: Type `Base` is not defined.
 class IssueInstanceTraceFrameAssoc(Base, PrepareMixin, RecordMixin):  # noqa
 
     __tablename__ = "issue_instance_trace_frame_assoc"
@@ -533,7 +283,6 @@ class SharedTextKind(enum.Enum):
         return cls.sink_detail
 
 
-# pyre-fixme[11]: Type `Base` is not defined.
 class SharedText(Base, PrepareMixin, RecordMixin):  # noqa
     """Any string-ish type that can be shared as a property of some other
     object. (e.g. features, sources, sinks). The table name 'messages' is due
@@ -543,12 +292,15 @@ class SharedText(Base, PrepareMixin, RecordMixin):  # noqa
 
     __table_args__ = (Index("ix_messages_handle", "contents", "kind"),)
 
+    # pyre-fixme[8]: Attribute has type `DBID`; used as `Column[typing.Any]`.
     id: DBID = Column(BIGDBIDType, primary_key=True)
 
+    # pyre-fixme[8]: Attribute has type `str`; used as `Column[str]`.
     contents: str = Column(
         String(length=SHARED_TEXT_LENGTH), nullable=False, index=True
     )
 
+    # pyre-fixme[8]: Attribute has type `SharedTextKind`; used as `Column[str]`.
     kind: SharedTextKind = Column(
         Enum(SharedTextKind), server_default="feature", nullable=False, index=True
     )
@@ -580,7 +332,6 @@ class SharedText(Base, PrepareMixin, RecordMixin):  # noqa
         )
 
 
-# pyre-fixme[11]: Type `Base` is not defined.
 class IssueInstanceSharedTextAssoc(Base, PrepareMixin, RecordMixin):  # noqa
     """Assoc table between issue instances and its properties that are
     representable by a string. The DB table name and column names are due to
@@ -646,12 +397,12 @@ class TraceKind(enum.Enum):
         return cls.postcondition
 
 
-# pyre-fixme[11]: Type `Base` is not defined.
 class IssueInstance(Base, PrepareMixin, MutableRecordMixin):  # noqa
     """A particularly instance of an issue found in a run"""
 
     __tablename__ = "issue_instances"
 
+    # pyre-fixme[8]: Attribute has type `DBID`; used as `Column[typing.Any]`.
     id: DBID = Column(BIGDBIDType, primary_key=True)
 
     location = Column(
@@ -770,6 +521,15 @@ class IssueInstance(Base, PrepareMixin, MutableRecordMixin):  # noqa
             i.is_new_issue = i.issue_id.is_new
             yield i
 
+    @classmethod
+    def generateSQLAlchemyObject(cls):
+        class IssueInstanceType(SQLAlchemyObjectType):
+            class Meta:
+                model = cls
+                interfaces = (relay.Node,)
+
+        return IssueInstanceType
+
 
 class IssueStatus(enum.Enum):
     """Issues are born uncategorized. Humans can
@@ -816,7 +576,6 @@ class IssueStatus(enum.Enum):
         return cls.do_not_care
 
 
-# pyre-fixme[11]: Type `Base` is not defined.
 class Issue(Base, PrepareMixin, MutableRecordMixin):  # noqa
     """An issue coming from the static analysis.
 
@@ -826,6 +585,7 @@ class Issue(Base, PrepareMixin, MutableRecordMixin):  # noqa
 
     __tablename__ = "issues"
 
+    # pyre-fixme[8]: Attribute has type `IssueDBID`; used as `Column[typing.Any]`.
     id: IssueDBID = Column(IssueBIGDBIDType, primary_key=True, nullable=False)
 
     handle = Column(
@@ -915,7 +675,6 @@ class RunStatus(enum.Enum):
 CURRENT_DB_VERSION = 1
 
 
-# pyre-fixme[11]: Type `Base` is not defined.
 class Run(Base):  # noqa
     """A particular run of the static analyzer.
 
@@ -1040,7 +799,6 @@ class Run(Base):  # noqa
         )
 
 
-# pyre-fixme[11]: Type `Base` is not defined.
 class MetaRun(Base):  # noqa
     """An identifier that represents multiple runs which should be grouped semantically.
 
@@ -1114,7 +872,6 @@ class RunSummary:
         return cls(**d)
 
 
-# pyre-fixme[11]: Type `Base` is not defined.
 class MetaRunToRunAssoc(Base, PrepareMixin, RecordMixin):  # noqa
     """The responsibility of filling out the meta-run to run assoc is on the child jobs
     of a larger run.
@@ -1140,7 +897,6 @@ class MetaRunToRunAssoc(Base, PrepareMixin, RecordMixin):  # noqa
         return cls._merge_assocs(session, items, cls.meta_run_id, cls.run_id)
 
 
-# pyre-fixme[11]: Type `Base` is not defined.
 class TraceFrameLeafAssoc(Base, PrepareMixin, RecordMixin):  # noqa
 
     __tablename__ = "trace_frame_message_assoc"
@@ -1173,10 +929,10 @@ class TraceFrameLeafAssoc(Base, PrepareMixin, RecordMixin):  # noqa
         return cls._merge_assocs(session, items, cls.trace_frame_id, cls.leaf_id)
 
 
-# pyre-fixme[11]: Type `Base` is not defined.
 class IssueInstanceFixInfo(Base, PrepareMixin, RecordMixin):  # noqa
     __tablename__ = "issue_instance_fix_info"
 
+    # pyre-fixme[8]: Attribute has type `DBID`; used as `Column[typing.Any]`.
     id: DBID = Column(BIGDBIDType, nullable=False, primary_key=True)
 
     fix_info = Column(String(length=INNODB_MAX_INDEX_LENGTH), nullable=False)
@@ -1190,7 +946,6 @@ class IssueInstanceFixInfo(Base, PrepareMixin, RecordMixin):  # noqa
     )
 
 
-# pyre-fixme[11]: Type `Base` is not defined.
 class TraceFrame(Base, PrepareMixin, RecordMixin):  # noqa
 
     __tablename__ = "trace_frames"
@@ -1200,6 +955,7 @@ class TraceFrame(Base, PrepareMixin, RecordMixin):  # noqa
         Index("ix_traceframe_run_callee_port", "run_id", "callee_id", "callee_port"),
     )
 
+    # pyre-fixme[8]: Attribute has type `DBID`; used as `Column[typing.Any]`.
     id: DBID = Column(BIGDBIDType, nullable=False, primary_key=True)
 
     kind = Column(Enum(TraceKind), nullable=False, index=False)
@@ -1212,6 +968,7 @@ class TraceFrame(Base, PrepareMixin, RecordMixin):  # noqa
         uselist=False,
     )
 
+    # pyre-fixme[8]: Attribute has type `str`; used as `Column[str]`.
     caller_port: str = Column(
         String(length=INNODB_MAX_INDEX_LENGTH),
         nullable=False,
@@ -1233,6 +990,7 @@ class TraceFrame(Base, PrepareMixin, RecordMixin):  # noqa
         doc="The location of the callee in the source code (line|start|end)",
     )
 
+    # pyre-fixme[8]: Attribute has type `str`; used as `Column[str]`.
     callee_port: str = Column(
         String(length=INNODB_MAX_INDEX_LENGTH),
         nullable=False,
@@ -1308,11 +1066,11 @@ class TraceFrame(Base, PrepareMixin, RecordMixin):  # noqa
 # This may be a message description, or it may be the start of another series
 # of traces leading to some other leaf. TraceFrameAnnotationTraceFrameAssoc
 # contains the first hop towards that leaf..
-# pyre-fixme[11]: Type `Base` is not defined.
 class TraceFrameAnnotation(Base, PrepareMixin, RecordMixin):  # noqa
 
     __tablename__ = "trace_frame_annotations"
 
+    # pyre-fixme[8]: Attribute has type `DBID`; used as `Column[typing.Any]`.
     id: DBID = Column(BIGDBIDType, nullable=False, primary_key=True)
 
     location = Column(
@@ -1321,6 +1079,7 @@ class TraceFrameAnnotation(Base, PrepareMixin, RecordMixin):  # noqa
 
     kind = Column(String(length=255), nullable=True, index=True)
 
+    # pyre-fixme[8]: Attribute has type `str`; used as `Column[str]`.
     message: str = Column(
         String(length=4096),
         doc="Message describing info about the trace",
@@ -1334,18 +1093,21 @@ class TraceFrameAnnotation(Base, PrepareMixin, RecordMixin):  # noqa
         uselist=False,
     )
 
+    # pyre-fixme[8]: Attribute has type `Optional[str]`; used as `Column[str]`.
     link: Optional[str] = Column(
         String(length=4096),
         doc="An optional URL linking the message to more info (Quandary)",
         nullable=True,
     )
 
+    # pyre-fixme[8]: Attribute has type `Optional[str]`; used as `Column[str]`.
     trace_key: Optional[str] = Column(
         String(length=INNODB_MAX_INDEX_LENGTH),
         nullable=True,
         doc="Link to possible pre/post traces (caller_condition).",
     )
 
+    # pyre-fixme[8]: Attribute has type `DBID`; used as `Column[typing.Any]`.
     trace_frame_id: DBID = Column(BIGDBIDType, nullable=False, index=True)
     trace_frame = relationship(
         "TraceFrame",
@@ -1372,7 +1134,6 @@ class TraceFrameAnnotation(Base, PrepareMixin, RecordMixin):  # noqa
 # the first hop trace frame from the annotation. It is similar to
 # IssueInstanceTraceFrameAssoc, which indicates the first hop trace frame from
 # the issue instance.
-# pyre-fixme[11]: Type `Base` is not defined.
 class TraceFrameAnnotationTraceFrameAssoc(Base, PrepareMixin, RecordMixin):  # noqa
 
     __tablename__ = "trace_frame_annotation_trace_frame_assoc"
@@ -1410,7 +1171,6 @@ class TraceFrameAnnotationTraceFrameAssoc(Base, PrepareMixin, RecordMixin):  # n
         )
 
 
-# pyre-fixme[11]: Type `Base` is not defined.
 class WarningMessage(Base):  # noqa
     __tablename__ = "warning_messages"
 
@@ -1436,7 +1196,6 @@ class WarningCodeCategory(enum.Enum):
         return cls.code_smell
 
 
-# pyre-fixme[11]: Type `Base` is not defined.
 class WarningCodeProperties(Base):  # noqa
     """Contains properties describing each warning code"""
 
@@ -1454,6 +1213,7 @@ class WarningCodeProperties(Base):  # noqa
         Enum(WarningCodeCategory),
         nullable=True,
         index=False,
+        # pyre-fixme[6]: Expected `str` for 4th param but got `Tuple[str]`.
         doc=(
             "The category of problems that issues in with this warning code "
             "can result in ",
@@ -1482,6 +1242,7 @@ class WarningCodeProperties(Base):  # noqa
         Float,
         nullable=True,
         index=False,
+        # pyre-fixme[6]: Expected `str` for 4th param but got `Tuple[str]`.
         doc=(
             "Average (min) length of forward traces for the given warning code "
             "(computed column)",
@@ -1492,6 +1253,7 @@ class WarningCodeProperties(Base):  # noqa
         Float,
         nullable=True,
         index=False,
+        # pyre-fixme[6]: Expected `str` for 4th param but got `Tuple[str]`.
         doc=(
             "Average (min) length of backward traces for the given warning "
             "code (computed column)",
@@ -1543,30 +1305,13 @@ class WarningCodeProperties(Base):  # noqa
     )
 
 
-# pyre-fixme[11]: Type `Base` is not defined.
-class PrimaryKey(Base, PrepareMixin, RecordMixin):  # noqa
-
-    __tablename__ = "primary_keys"
-
-    table_name: str = Column(
-        String(length=100),
-        doc="Name of the table that this row stores the next available primary key for",
-        nullable=False,
-        primary_key=True,
-    )
-
-    current_id: int = Column(
-        BIGINT(unsigned=True).with_variant(BIGINT, "sqlite"),
-        doc="The current/latest id used in the table.",
-        nullable=False,
-        primary_key=False,
-    )
+class PrimaryKey(Base, PrimaryKeyBase):
+    pass
 
 
-class PrimaryKeyGenerator:
-    """Keep track of DB objects' primary keys by ourselves rather than relying
-    on SQLAlchemy, so we can supply them as arguments when creating association
-    objects, such as TraceFrameLeafAssoc"""
+class PrimaryKeyGenerator(PrimaryKeyGeneratorBase):
+
+    PRIMARY_KEY: Type = PrimaryKey
 
     QUERY_CLASSES: Set[Type] = {
         Issue,
@@ -1578,102 +1323,9 @@ class PrimaryKeyGenerator:
         TraceFrameAnnotation,
     }
 
-    # Map from class name to an ID range (next_id, max_reserved_id)
-    pks: Dict[str, Tuple[int, int]] = {}
 
-    def reserve(
-        self,
-        session: Session,
-        saving_classes: List[Type],
-        item_counts: Optional[Dict[str, int]] = None,
-        use_lock: bool = False,
-    ) -> "PrimaryKeyGenerator":
-        """
-        session - Session for DB operations.
-        saving_classes - class objects that need to be saved e.g. Issue, Run
-        item_counts - map from class name to the number of items, for preallocating
-        id ranges
-        """
-        query_classes = {cls for cls in saving_classes if cls in self.QUERY_CLASSES}
-        for cls in query_classes:
-            if item_counts and cls.__name__ in item_counts:
-                count = item_counts[cls.__name__]
-            else:
-                count = 1
-            self._reserve_id_range(session, cls, count, use_lock)
-
-        return self
-
-    def _lock_pk_with_retries(
-        self, session: Session, cls: Type
-    ) -> Optional[PrimaryKey]:
-        cls_pk: Optional[PrimaryKey] = None
-        retries: int = 3
-        while retries > 0:
-            try:
-                cls_pk = (
-                    session.query(PrimaryKey)
-                    .filter(PrimaryKey.table_name == cls.__name__)
-                    .with_for_update()
-                    .first()
-                )
-                # if we're here, the record has been locked, or there is no record
-                retries = 0
-            except exc.OperationalError as ex:
-                # Failed to get exclusive lock on the record, so we retry
-                retries -= 1
-                # Re-raise the exception if our retries are exhausted
-                if retries == 0:
-                    raise ex
-        return cls_pk
-
-    def _reserve_id_range(
-        self, session: Session, cls: Type, count: int, use_lock: bool
-    ) -> None:
-        cls_pk = self._lock_pk_with_retries(session, cls)
-        if use_lock or not cls_pk:
-            # If cls_pk is None, then we query the data table for the max ID
-            # and use that as the current_id in the primary_keys table. This
-            # should only occur once (the except with a rollback means any
-            # additional attempt will fail to add a row, and use the "current"
-            # id value)
-            row = session.query(cls.id).order_by(cls.id.desc()).first()
-            try:
-                if not cls_pk:
-                    session.execute(
-                        "INSERT INTO primary_keys(table_name, current_id) \
-                        VALUES (:table_name, :current_id)",
-                        {
-                            "table_name": cls.__name__,
-                            "current_id": (row.id) if row else 0,
-                        },
-                    )
-                else:
-                    cls_pk.current_id = row.id if row else 0
-                session.commit()
-            except exc.SQLAlchemyError:
-                session.rollback()
-            cls_pk = self._lock_pk_with_retries(session, cls)
-
-        if cls_pk:
-            next_id = cls_pk.current_id + 1
-            cls_pk.current_id = cls_pk.current_id + count
-            pk_entry: Tuple[int, int] = (next_id, cls_pk.current_id)
-            session.commit()
-            self.pks[cls.__name__] = pk_entry
-
-    def get(self, cls):
-        assert cls in self.QUERY_CLASSES, (
-            "%s primary key should be generated by SQLAlchemy" % cls.__name__
-        )
-        assert cls.__name__ in self.pks, (
-            "%s primary key needs to be initialized before use" % cls.__name__
-        )
-        (pk, max_pk) = self.pks[cls.__name__]
-        assert pk <= max_pk, "%s reserved primary key range exhausted" % cls.__name__
-        self.pks[cls.__name__] = (pk + 1, max_pk)
-        return pk
-
-
-def create(engine):
-    Base.metadata.create_all(engine)
+def create(db: DB):
+    try:
+        Base.metadata.create_all(db.engine)  # pyre-ignore [16]
+    except NoSuchTableError:
+        pass
